'''
Created on 2022-03-24

@author: wf
'''

class Version(object):
    '''
    Version handling for py-3rdparty-mediawiki
    '''
    name="py-3rdparty-mediawiki"
    version='0.7.2'
    date = '2020-10-31'
    updated = '2022-11-24'